/*
 * assembler.c: Read the VM-2 Assembly code specification of a program
 *              and produce the executable code in binary or hexadecimal
 *              form.
 *              The syntax of the VM-2 Assembly language is the following.
 *              Keywords are in Capital letters only.
 *              Identifiers start with lower case letters only.
 *              Integers are coded in decimal form.
 *              Spaces, tabs and end-of-line can be inserted in arbitrary way.
 *              Single line comments start with a "#" sign and end with the
 *              current line; they may contain any sequence of ASCII
 *              characters except end-of-line.
 *              Multiple line comments start with a "/" followed by a "*"
 *              with no extra blanks in between, and terminate with
 *              "*" followed by "/" with no extra blanks in between.
 *           
 *     <program>     ::= <instruct> ; <program>
 *     <program>     ::= <instruct> .
 *     <instruct>    ::= <label> : <instru>
 *     <instruct>    ::= <label> : <int16>
 *     <instruct>    ::= <instru>
 *     <label>       ::= <ident>
 *     <label>       ::= <int12>
 *     <instru>      ::= <instrd>
 *     <instru>      ::= <instrl>
 *     <instru>      ::= <instri>
 *     <instru>      ::= <instr8>
 *     <instru>      ::= <instr0>
 *     <instru>      ::= <instrs>
 *     <instrd>      ::= JPOS <label>
 *     <instrd>      ::= JNEG <label>
 *     <instrd>      ::= JZER <label>
 *     <instrd>      ::= JNZE <label>
 *     <instrd>      ::= LODD <label>
 *     <instrd>      ::= ADDD <label>
 *     <instrd>      ::= ANDD <label>
 *     <instrd>      ::= STOD <label>
 *     <instrd>      ::= CALL <label>
 *     <instrd>      ::= JUMP <label>
 *     <instrl>      ::= LODL <label>
 *     <instrl>      ::= ADDL <label>
 *     <instrl>      ::= ANDL <label>
 *     <instrl>      ::= STOL <label>
 *     <instri>      ::= LDIX <int12>
 *     <instr8>      ::= LOC8 <int8>
 *     <instr8>      ::= LOCH <int8>
 *     <instr8>      ::= INSP <int8>
 *     <instr8>      ::= DESP <int8>
 *     <instr0>      ::= POP
 *     <instr0>      ::= POPI
 *     <instr0>      ::= SWAS
 *     <instr0>      ::= SWAF
 *     <instr0>      ::= NEGA
 *     <instr0>      ::= SIGN
 *     <instr0>      ::= RETN
 *     <instr0>      ::= HALT
 *     <instr0>      ::= RSHF
 *     <instr0>      ::= LSHF
 *     <instr0>      ::= PUSH
 *     <instr0>      ::= PSHI
 *     <instrs>      ::= DEF <ident> = <int12>
 *
 * Copyright C 2000 by Giovanni Chiola <chiola@disi.unige.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Release: 1.0, 26/June/2000
 * Modified by G. Ciaccio 4 dic 2000 (see at end)
 *
 */

#include <stdio.h>
#include <stdlib.h>

#ifndef TRUE
#  define TRUE 1
#  define FALSE 0
#endif TRUE

#define opcode_JPOS 0x0000
#define opcode_JNEG 0x1000
#define opcode_JZER 0x2000
#define opcode_JNZE 0x3000
#define opcode_LODD 0x4000
#define opcode_ADDD 0x5000
#define opcode_ANDD 0x6000
#define opcode_STOD 0x7000
#define opcode_LODL 0x8000
#define opcode_ADDL 0x9000
#define opcode_ANDL 0xA000
#define opcode_STOL 0xB000
#define opcode_JUMP 0xC000
#define opcode_LDIX 0xD000
#define opcode_CALL 0xE000
#define opcode_LOC8 0xF000
#define opcode_LOCH 0xF100
#define opcode_INSP 0xF200
#define opcode_DESP 0xF300
#define opcode_PUSH 0xF400
#define opcode_PSHI 0xF500
#define opcode_POP  0xF600
#define opcode_POPI 0xF700
#define opcode_SWAS 0xF800
#define opcode_SWAF 0xF900
#define opcode_NEGA 0xFA00
#define opcode_SIGN 0xFB00
#define opcode_RETN 0xFC00
#define opcode_HALT 0xFD00
#define opcode_RSHF 0xFE00
#define opcode_LSHF 0xFF00

#define Identif     -1
#define Number      -2
#define EndLabel    -3
#define Assign      -4
#define InstrSepar  -5
#define DefDecl     -6
#define EndProgr    -7

#define VM2_RAM_Size 4094
#define MAX_DEF 1000

typedef struct def_str {
  char * nam;
  unsigned val;
} Def_label;

typedef struct und_str {
  char * nam;
  int fst;
} UnD_label;

static unsigned RAM[VM2_RAM_Size];
static int undef_link[VM2_RAM_Size];
static unsigned undef_mask[VM2_RAM_Size];
static char * Rlabels[VM2_RAM_Size];
static unsigned char cell_used[VM2_RAM_Size];

static Def_label identifier[MAX_DEF];
static UnD_label unidentif[MAX_DEF];
static unsigned no_ident;
static unsigned no_unid;

static unsigned cur_pc;
static unsigned max_pc;

void init(void){
  int i;
  for ( i = 0 ; i < VM2_RAM_Size ; i++ ) {
      RAM[i] = 0; cell_used[i] = FALSE;
      Rlabels[i] = NULL;
      undef_mask[i] = 0;
      undef_link[i] = -1;
    }
  cur_pc = max_pc = no_ident = no_unid = 0;
}

char * my_str_alloc(const char * s) {
  char * my_str;
  size_t len;
  if ( (my_str = (char *)malloc(len = strlen(s)+1)) == NULL ) {
      fprintf(stdout,"\nmy_str_alloc ERROR: couldnt allocate %1d bytes !\n\n",
              (int)len);
      exit(1);
    }
  return strcpy(my_str,s);
}

static char my_string[1000];
static unsigned my_number;

inline void syntax_error(char * my_err) {
    fprintf(stdout,"\nSYNTAX ERROR: %s ! Exiting\n", my_err);
    exit(1);
}

static int next_lex_term(FILE * in_f, FILE * out_f) {
    char nc, c2, pc;
    int i;

      pc = ' ';
      while ( ((nc = getc(in_f)) == '#') || (nc == ' ') || (nc == '\t') ||
              (nc == '\n') || (nc == '/') || ((nc == '*')&&(pc == '/')) ) {
          putc(nc,out_f);
          pc = nc;
          if ( nc == '#' ) {
              while ( (c2 = getc(in_f)) != '\n' )
                  putc(c2,out_f);
              putc('\n',out_f);
            }
          else if ( nc == '*' ) {
              for ( pc = ' ', c2 = getc(in_f); (pc  != '*')&&(c2 != '/') ;
                    pc = c2, c2 = getc(in_f) ) {
                  putc(c2,out_f);
                }
              putc('/',out_f);
            }
        }
      if ( nc >= 'a' && nc <= 'z' ) /* Identifier */ {
          putc(nc,out_f);
          my_string[0] = nc; i = 1;
          while ( (((c2 = getc(in_f)) >= 'a') && (c2 <= 'z'))
                  || (c2 == '_') || ((c2 >= '0') && (c2 <= '9'))
                  || ((c2 >= 'A') && (c2 <= 'Z')) ) {
              my_string[i] = c2; i++;
              putc(c2,out_f);
            }
          ungetc(c2,in_f); my_string[i] = '\0';
          return Identif;
      } else if ( nc >= '0' && nc <= '9' ) /* Number */ {
          ungetc(nc,in_f);
          fscanf(in_f,"%d",&my_number);
          fprintf(out_f,"%1d",my_number);
          return Number;
      } else switch (nc) {
        case '-': fscanf(in_f,"%d",&my_number);
                  fprintf(out_f,"-%1d",my_number);
                  my_number ^= 0xFFFF; my_number++;
                  my_number &= 0xFFFF;
                  return Number;
                 break;
        case ':': putc(nc,out_f);
                  return EndLabel;
                 break;
        case '=': putc(nc,out_f);
                  return Assign;
                 break;
        case ';': putc(nc,out_f);
                  return InstrSepar;
                 break;
        case '.': putc(nc,out_f);
                  return EndProgr;
                 break;
        case 'A': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'D' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'D' )
                          syntax_error("D expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 == 'D' )
                          return opcode_ADDD;
                      if ( c2 == 'L' )
                          return opcode_ADDL;
                      syntax_error("D or L expected");
                  } else if ( c2 == 'N' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'D' )
                          syntax_error("D expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 == 'D' )
                          return opcode_ANDD;
                      if ( c2 == 'L' )
                          return opcode_ANDL;
                      syntax_error("D or L expected");
                  } else
                      syntax_error("D or N expected");
                 break;
        case 'C': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'A' )
                      syntax_error("A expected");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'L' )
                      syntax_error("L expected");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'L' )
                      syntax_error("L expected");
                  return opcode_CALL;
                 break;
        case 'D': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'E' )
                      syntax_error("E expected");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'F' )
                      return DefDecl;
                  if ( c2 != 'S' )
                      syntax_error("F or S expected");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'P' )
                      syntax_error("P expected");
                  return opcode_DESP;
                 break;
        case 'H': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'A' )
                      syntax_error("A expected");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'L' )
                      syntax_error("L expected");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'T' )
                      syntax_error("T expected");
                  return opcode_HALT;
                 break;
        case 'I': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'N' )
                      syntax_error("N expected");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'S' )
                      syntax_error("S expected");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'P' )
                      syntax_error("P expected");
                  return opcode_INSP;
                 break;
        case 'J': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'N' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 == 'E' ) {
                          c2 = getc(in_f); putc(c2,out_f);
                          if ( c2 != 'G' )
                              syntax_error("G expected");
                          return opcode_JNEG;
                      } else if ( c2 == 'Z' ) {
                          c2 = getc(in_f); putc(c2,out_f);
                          if ( c2 != 'E' )
                              syntax_error("E expected");
                          return opcode_JNZE;
                      } else
                          syntax_error("E or Z expected");
                  } else if ( c2 == 'P' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'O' )
                          syntax_error("O expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'S' )
                          syntax_error("S expected");
                      return opcode_JPOS;
                  } else if ( c2 == 'U' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'M' )
                          syntax_error("M expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'P' )
                          syntax_error("P expected");
                      return opcode_JUMP;
                  } else if ( c2 == 'Z' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'E' )
                          syntax_error("E expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'R' )
                          syntax_error("R expected");
                      return opcode_JZER;
                  } else
                      syntax_error("N or P or U or Z expected");
                 break;
        case 'L': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'D' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'I' )
                          syntax_error("I expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'X' )
                          syntax_error("X expected");
                      return opcode_LDIX;
                  } else if ( c2 == 'O' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 == 'C' ) {
                          c2 = getc(in_f); putc(c2,out_f);
                          if ( c2 == '8' )
                              return opcode_LOC8;
                          else if ( c2 == 'H' )
                              return opcode_LOCH;
                          else
                              syntax_error("8 or H expected");
                      } else if ( c2 == 'D' ) {
                          c2 = getc(in_f); putc(c2,out_f);
                          if ( c2 == 'D' )
                              return opcode_LODD;
                          else if ( c2 == 'L' )
                              return opcode_LODL;
                          else
                              syntax_error("D or L expected");
                      } else
                          syntax_error("D or L expected");
                  } else if ( c2 == 'S' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'H' )
                          syntax_error("H expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'F' )
                          syntax_error("F expected");
                      return opcode_LSHF;
                  } else
                      syntax_error("D or O or S expected");
                 break;
        case 'N': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'E' )
                      syntax_error("E expected");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'G' )
                      syntax_error("G expected");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'A' )
                      syntax_error("A expected");
                  return opcode_NEGA;
                 break;
        case 'P': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'O' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'P' )
                          syntax_error("P expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 == 'I' )
                          return opcode_POPI;
                      else {
                          ungetc(c2,in_f);
                          return opcode_POP;
                        }
                  } else if ( c2 == 'S' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'H' )
                          syntax_error("H expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'I' )
                          syntax_error("I expected");
                      return opcode_PSHI;
                  } else if ( c2 == 'U' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'S' )
                          syntax_error("S expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'H' )
                          syntax_error("H expected");
                      return opcode_PUSH;
                  } else
                      syntax_error("O or S or U expected");
                 break;
        case 'R': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'E' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'T' )
                          syntax_error("T expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'N' )
                          syntax_error("N expected");
                      return opcode_RETN;
                  } else if ( c2 == 'S' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'H' )
                          syntax_error("H expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'F' )
                          syntax_error("F expected");
                      return opcode_RSHF;
                  } else
                      syntax_error("E or S expected");
                 break;
        case 'S': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'I' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'G' )
                          syntax_error("G expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'N' )
                          syntax_error("N expected");
                      return opcode_SIGN;
                  } else if ( c2 == 'T' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'O' )
                          syntax_error("O expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 == 'D' ) {
                          return opcode_STOD;
                      } else if ( c2 == 'L' ) {
                          return opcode_STOL;
                      } else
                          syntax_error("D or L expected");
                  } else if ( c2 == 'W' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'A' )
                          syntax_error("A expected");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 == 'F' ) {
                          return opcode_SWAF;
                      } else if ( c2 == 'S' ) {
                          return opcode_SWAS;
                      } else
                          syntax_error("F or S expected");
                  } else
                      syntax_error("I or T or W expected");
                 break;
        default : putc(nc,out_f);
                  fprintf(out_f,"\n??? LEXICAL ERROR !!!\n");
                exit(1);
      }
      return 0;
}

static void read_instruct(FILE * in_f, FILE * out_f) {
    int i, j;
    int cut, nxt;

    cut = next_lex_term(in_f,out_f);
    if ( cut == Number || cut == Identif ) {
        if ( (j = next_lex_term(in_f,out_f)) != EndLabel )
            syntax_error("colon expected");
        if ( cut == Number ) {
            if ( my_number >= VM2_RAM_Size ) {
                fprintf(out_f,"\n*** ERROR: label value=%1d too high!\n\n",
                        my_number);
                exit(1);
              }
            if ( cell_used[my_number] ) {
                fprintf(out_f,"\n*** ERROR: location %1d already used!\n\n",
                        my_number);
                exit(1);
              }
            cur_pc = my_number;
        } else { /* if ( cut == Identif ) */
            for ( j = 0 ; j <= max_pc ; j++ )
                if ( (Rlabels[j] != NULL) &&
                     (strcmp(Rlabels[j],my_string) == 0) ) {
                    fprintf(out_f,
                    "\n*** ERROR: label %s is defined more than once!\n\n",
                            my_string);
                    exit(1);
                  }
            if ( Rlabels[cur_pc] != NULL ) {
                fprintf(out_f,
  "\n*** ERROR: another label (%s) is already defined for location %1d!\n\n",
                        my_string, cur_pc);
                exit(1);
              }
            Rlabels[cur_pc] = my_str_alloc(my_string);
            for ( j = 0 ; j < no_unid ; j++ )
                if ( strcmp(unidentif[j].nam,my_string) == 0 ) {
                    for ( i = unidentif[j].fst; i >= 0 ; i = undef_link[i] ) {
                        RAM[i] |= (cur_pc & undef_mask[i]);
                      }
                    i = (no_unid-1);
                    if ( j < i )
                        unidentif[j] = unidentif[i];
                    no_unid = i;
                    break;
                  }
          }
        cut = next_lex_term(in_f,out_f);
      }
    if ( cut == DefDecl ) {
        if ( (nxt = next_lex_term(in_f,out_f)) != Identif )
            syntax_error("Identifier expected");
        for ( j = 0 ; j <= max_pc ; j++ )
            if ( (Rlabels[j] != NULL) &&
                 (strcmp(Rlabels[j],my_string) == 0) ) {
                fprintf(out_f,
        "\n*** ERROR: label %s is already associated with location %1d !\n\n",
                        my_string, j);
                exit(1);
              }
        for ( j = 0 ; j < no_ident ; j++ )
            if ( (strcmp(identifier[j].nam,my_string) == 0) ) {
                fprintf(out_f,
                "\n*** ERROR: identifier %s is already declared !\n\n",
                        my_string);
                exit(1);
              }
        identifier[no_ident].nam = my_str_alloc(my_string);
        if ( (nxt = next_lex_term(in_f,out_f)) != Assign )
            syntax_error("equal sign expected");
        if ( (nxt = next_lex_term(in_f,out_f)) != Number )
            syntax_error("number expected");
        identifier[no_ident].val = (my_number & 0xFFF);
        no_ident++;
        for ( j = 0 ; j < no_unid ; j++ )
            if ( strcmp(unidentif[j].nam,my_string) == 0 ) {
                for ( i = unidentif[j].fst; i >= 0 ; i = undef_link[i] ) {
                    RAM[i] |= (my_number & undef_mask[i]);
                  }
                i = (no_unid-1);
                if ( j < i )
                    unidentif[j] = unidentif[i];
                no_unid = i;
                break;
              }
        return;
      }
    if ( cur_pc > max_pc )
        max_pc = cur_pc;
    cell_used[cur_pc] = TRUE;
    if ( cut == Number ) {
        if ( my_number > 0xFFFF ) {
            fprintf(out_f,
            "\n*** ERROR: value %1d is too high to fit in location %1d!\n\n",
                    my_number, cur_pc);
            exit(1);
          }
        RAM[cur_pc] = my_number;
        cur_pc++;
        return;
      }
    switch (cut) {
        case opcode_JPOS:
        case opcode_JNEG:
        case opcode_JZER:
        case opcode_JNZE:
        case opcode_LODD:
        case opcode_ADDD:
        case opcode_ANDD:
        case opcode_STOD:
        case opcode_JUMP:
        case opcode_CALL:
            if ( ((nxt = next_lex_term(in_f,out_f)) != Number) &&
                 (nxt != Identif) )
                syntax_error("number or identifier expected");
            if ( nxt == Number )
                RAM[cur_pc] = (my_number & 0xFFF);
            else {
              int search = TRUE;
                RAM[cur_pc] = 0;
                for ( j = 0 ; j <= max_pc ; j++ )
                    if ( (Rlabels[j] != NULL) &&
                         (strcmp(Rlabels[j],my_string) == 0) ) {
                        RAM[cur_pc] = j;
                        search = FALSE;
                        break;
                      }
                for ( j = 0 ; search && (j < no_ident) ; j++ )
                    if ( (strcmp(identifier[j].nam,my_string) == 0) ) {
                        RAM[cur_pc] = identifier[j].val;
                        search = FALSE;
                      }
                if ( search ) { /* add to to-do list */
                    for ( j = 0 ; j < no_unid ; j++ )
                        if ( strcmp(unidentif[j].nam,my_string) == 0 ) {
                            goto found_1;
                          }
                    /* not found */
                    j = no_unid;
                    unidentif[j].nam = my_str_alloc(my_string);
                    unidentif[j].fst = -1;
                    no_unid++;
                  found_1:
                    undef_link[cur_pc] = unidentif[j].fst;
                    unidentif[j].fst = cur_pc;
                    undef_mask[cur_pc] = 0xFFF;
                  }
              }
          break;
        case opcode_LODL:
        case opcode_ADDL:
        case opcode_ANDL:
        case opcode_STOL:
        case opcode_LDIX:
            if ( ((nxt = next_lex_term(in_f,out_f)) != Number) &&
                 (nxt != Identif) )
                syntax_error("number or identifier expected");
            if ( nxt == Number )
                RAM[cur_pc] = (my_number & 0xFFF);
            else {
              int search = TRUE;
                RAM[cur_pc] = 0;
                for ( j = 0 ; search && (j < no_ident) ; j++ )
                    if ( (strcmp(identifier[j].nam,my_string) == 0) ) {
                        RAM[cur_pc] = identifier[j].val;
                        search = FALSE;
                      }
                if ( search ) { /* add to to-do list */
                    for ( j = 0 ; j < no_unid ; j++ )
                        if ( strcmp(unidentif[j].nam,my_string) == 0 ) {
                            goto found_2;
                          }
                    /* not found */
                    j = no_unid;
                    unidentif[j].nam = my_str_alloc(my_string);
                    unidentif[j].fst = -1;
                    no_unid++;
                  found_2:
                    undef_link[cur_pc] = unidentif[j].fst;
                    unidentif[j].fst = cur_pc;
                    undef_mask[cur_pc] = 0xFFF;
                  }
              }
          break;
        case opcode_LOC8:
        case opcode_LOCH:
        case opcode_INSP:
        case opcode_DESP:
            if ( ((nxt = next_lex_term(in_f,out_f)) != Number) &&
                 (nxt != Identif) )
                syntax_error("number or identifier expected");
            if ( nxt == Number ) {
                RAM[cur_pc] = (my_number & 0xFF);
              }
            else {
              int search = TRUE;
                RAM[cur_pc] = 0;
                for ( j = 0 ; search && (j < no_ident) ; j++ )
                    if ( (strcmp(identifier[j].nam,my_string) == 0) ) {
                        RAM[cur_pc] = (0xFF & (identifier[j].val));
                        search = FALSE;
                      }
                if ( search ) { /* add to to-do list */
                    for ( j = 0 ; j < no_unid ; j++ )
                        if ( strcmp(unidentif[j].nam,my_string) == 0 ) {
                            goto found_3;
                          }
                    /* not found */
                    j = no_unid;
                    unidentif[j].nam = my_str_alloc(my_string);
                    unidentif[j].fst = -1;
                    no_unid++;
                  found_3:
                    undef_link[cur_pc] = unidentif[j].fst;
                    unidentif[j].fst = cur_pc;
                    undef_mask[cur_pc] = 0xFF;
                  }
              }
          break;
        case opcode_POP :
        case opcode_POPI:
        case opcode_SWAS:
        case opcode_SWAF:
        case opcode_NEGA:
        case opcode_SIGN:
        case opcode_RETN:
        case opcode_HALT:
        case opcode_RSHF:
        case opcode_LSHF:
        case opcode_PUSH:
        case opcode_PSHI:
            RAM[cur_pc] = 0;
          break;
        default:
            syntax_error("instruction or number or DEF expected");
        }
    RAM[cur_pc] |= cut;
    cur_pc++;
}


/*
 *  Names of input and output files can now be typed as line args.
 *  G. Ciaccio  4 dec 2000
 */
void main(int argc, char **argv) {
    char inp_f_nam[1000];
    char out_f_nam[1000];
    FILE * infp, * oufp;
    int nxt, j;

    printf(
"VM-2 Assembler, Copyright C 2000 by Giovanni Chiola <chiola@disi.unige.it>\n"
            );
    if (argc > 1)
        strcpy(inp_f_nam,argv[1]);
    else {
         printf("Input file name:"); fflush(stdout);
         scanf("%s",inp_f_nam);
    }
    infp = fopen(inp_f_nam,"r");
    if ( infp == NULL ) {
        printf("\n*** ERROR: couldnt open input file %s\n\n",inp_f_nam);
        exit(1);
      }
    if (argc > 2)
        strcpy(out_f_nam,argv[2]);
    else {
         printf("Output file name:"); fflush(stdout);
         scanf("%s",out_f_nam);
    }
    oufp = fopen(out_f_nam,"w");
    if ( oufp == NULL ) {
        printf("\n*** ERROR: couldnt open output file %s\n\n",out_f_nam);
        exit(1);
      }
    init();
    do {
        read_instruct(infp,stdout);
        nxt = next_lex_term(infp,stdout);
    } while ( nxt == InstrSepar );
    if ( nxt != EndProgr )
        syntax_error("semicolon or fullstop expected");
    if ( no_unid > 0 ) {
        printf("\n*** ERROR: the following identifiers are undefined:\n");
        while ( no_unid > 0 ) {
            no_unid--;
            printf("  %s (used in the following locations:",
                  unidentif[no_unid].nam);
            for ( j = unidentif[no_unid].fst; j >= 0 ; j = undef_link[j] ) {
                printf(" %1d", j);
              }
            printf("\n");
          }
        exit(1);
      }
    for ( j = 0 ; j <= max_pc ; j++ )
        fprintf(oufp,"%4x\n",RAM[j]);
/*
 *  Add a blank line at the end -- it seems necessary for the simulator to
 *  work.
 *  Actually it is the simulator that needs be bug-fixed, not the assembler :)
 *  G. Ciaccio 4 dic 2000 
 */
    fprintf(oufp,"\n");
}

